/**
 * @file FdSerialTest.c
 * Test file for FdSerial.c module
 *
 * Copyright (c) 2008, Steve Denson
 * See end of file for terms of use.
 */
#include <stdio.h>
#include "FdSerial.h"

/*
 * get a character from serial interface
 */
int getchar(void)
{
    return FdSerial_rx();
}
/*
 * check for character input without blocking
 */
int getkey(void)
{
    return FdSerial_rxcheck();
}

/*
 * putchar must be defined for printf to work
 */
int putchar(char ch)
{
    if(ch == '\n') FdSerial_tx('\r');
    FdSerial_tx(ch);
    return ch;
}

/*
 * Using main without args for embedded use.
 * defining FDS_DISABLE_OUTS saves ~500 bytes
 * FdSerial_bin FdSerial_dec FdSerial_hex FdSerial_str etc...
 * are not necessary because we have printf.
 */
void main(void)
{
    int ii = 0;
    int hex= 0x1deadbee;

    FdSerial_start(31,30,0,115200);
    FdSerial_rx();
    printf("\nHello World!\n");
#ifndef FDS_DISABLE_OUTS	
    FdSerial_hex(hex,8);
    putchar(' ');
    FdSerial_dec(hex);
    putchar('\n');
#endif	
    printf("%x hex is %d decimal\n", hex, hex);
    while(1) {
        ii = getkey();
        if(ii >= 0)
            putchar(ii);
    }
}

/*
+------------------------------------------------------------------------------------------------------------------------------+
                                                   TERMS OF USE: MIT License                                                                                                              
+------------------------------------------------------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation     
files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    
modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:                                                                   
                                                                                                                              
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
                                                                                                                              
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         
+------------------------------------------------------------------------------------------------------------------------------+
*/